package ch.psi.daq.queryrest.response.csv;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.ServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.supercsv.cellprocessor.constraint.NotNull;
import org.supercsv.cellprocessor.ift.CellProcessor;
import org.supercsv.io.dozer.CsvDozerBeanWriter;
import org.supercsv.io.dozer.ICsvDozerBeanWriter;
import org.supercsv.prefs.CsvPreference;

import com.fasterxml.jackson.core.JsonEncoding;

import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.impl.DAQQuery;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class CSVResponseStreamWriter implements ResponseStreamWriter {
   public static final String APPLICATION_CSV_VALUE = "text/csv";

   private static final char DELIMITER_CVS = ';';
   private static final char DELIMITER_ARRAY = ',';

   private static final Logger LOGGER = LoggerFactory.getLogger(CSVResponseStreamWriter.class);
   
   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   @Override
   public void respond(Stream<Entry<String, ?>> stream, DAQQuery query, ServletResponse response) throws Exception {
      response.setCharacterEncoding(JsonEncoding.UTF8.getJavaName());
      response.setContentType(APPLICATION_CSV_VALUE);

      respondInternal(stream, query, response);
   }

   private void respondInternal(Stream<Entry<String, ?>> stream, DAQQuery query, ServletResponse response)
         throws Exception {
      Set<QueryField> queryFields = query.getFields();
      List<Aggregation> aggregations = query.getAggregations();

      Set<String> fieldMapping =
            new LinkedHashSet<>(queryFields.size() + (aggregations != null ? aggregations.size() : 0));
      List<String> header =
            new ArrayList<>(queryFields.size() + (aggregations != null ? aggregations.size() : 0));
      List<CellProcessor> processorSet =
            new ArrayList<>(queryFields.size() + (aggregations != null ? aggregations.size() : 0));
      boolean isNewField;
      QueryAnalyzer queryAnalyzer = queryAnalizerFactory.apply(query);

      for (QueryField field : queryFields) {
         if(!(QueryField.value.equals(field) && queryAnalyzer.isAggregationEnabled())){
            isNewField = fieldMapping.add(field.name());

            if (isNewField) {
               header.add(field.name());
               processorSet.add(new ArrayProcessor(new NotNull(), DELIMITER_ARRAY));
            }
         }
      }
      if (aggregations != null && queryAnalyzer.isAggregationEnabled()) {
         for (Aggregation aggregation : query.getAggregations()) {
            isNewField = fieldMapping.add("value." + aggregation.name());

            if (isNewField) {
               header.add(aggregation.name());
               processorSet.add(new ArrayProcessor(new NotNull(), DELIMITER_ARRAY));
            }
         }
      }

      CellProcessor[] processors = processorSet.toArray(new CellProcessor[processorSet.size()]);
      CsvPreference preference = new CsvPreference.Builder(
            (char) CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE.getQuoteChar(),
            DELIMITER_CVS,
            CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE.getEndOfLineSymbols()).build();
      ICsvDozerBeanWriter beanWriter = new CsvDozerBeanWriter(response.getWriter(), preference);

      AtomicReference<Exception> exception = new AtomicReference<>();
      try {
         // configure the mapping from the fields to the CSV columns
         beanWriter.configureBeanMapping(DataEvent.class, fieldMapping.toArray(new String[fieldMapping.size()]));

         beanWriter.writeHeader(header.toArray(new String[header.size()]));

         stream
               /* ensure elements are sequentially written */
               .sequential()
               .forEach(
                     entry -> {
                        if (entry.getValue() instanceof Stream) {

                           @SuppressWarnings("unchecked")
                           Stream<DataEvent> eventStream = (Stream<DataEvent>) entry.getValue();
                           eventStream
                                 .forEach(
                                 event -> {
                                    try {
                                       beanWriter.write(event, processors);
                                    } catch (Exception e) {
                                       LOGGER.error("Could not write elements for '{}-{}'", event.getChannel(),
                                             event.getPulseId(), e);
                                       exception.compareAndSet(null, e);
                                    }
                                 });
                        } else {
                           String message = "Type '" + entry.getValue().getClass() + "' not supported.";
                           LOGGER.error(message);
                           exception.compareAndSet(null, new RuntimeException(message));
                        }
                     });
      } finally {
         beanWriter.close();
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

}
